from collections import OrderedDict
from datetime import datetime
from logging import Logger
from typing import Dict, List, Optional

from rich import box
from rich.align import Align
from rich.console import RenderGroup
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree


class MessagePanel:
    def __init__(self, topic: str, message_buffer: "OrderedDict[float, str]"):
        self._message_buffer = message_buffer
        self._topic = topic
        self._last_timestamp: Optional[str] = None
        self._last_value: Optional[str] = None

    def __rich__(self) -> Panel:
        table = Table(
            expand=True,
            header_style="bold blue",
            border_style="white",
            highlight=True,
            style="white",
        )
        table.add_column("Timestamp", justify="left")
        table.add_column("Message", justify="left")

        if self._message_buffer:
            items = self._message_buffer.popitem(last=False)
            self._last_timestamp, self._last_value = (
                str(datetime.fromtimestamp(items[0])),
                str(items[1]),
            )

        if self._last_timestamp and self._last_value:
            table.add_row(self._last_timestamp, self._last_value)

        return Panel(
            Align.center(
                RenderGroup(table),
                vertical="middle",
            ),
            box=box.ROUNDED,
            padding=(1, 2),
            title=f"[b blue] Messages from Topic: {self._topic}",
            border_style="bright_blue",
        )


class Header:
    def __rich__(self) -> Panel:
        table = Table.grid(expand=True)
        table.add_column(justify="center", ratio=1)
        table.add_column(justify="left")
        table.add_row(
            "[b]ESS console-message-consumer[/b]",
            datetime.now().ctime().replace(":", "[blink]:[/]"),
        )
        return Panel(table, style="white on blue")


class TopicsTreePanel:
    def __init__(self, existing_topics: List[str]):
        self._existing_topics = existing_topics

    def __rich__(self) -> Panel:
        tree = Tree(
            "Topics",
            guide_style="bright_blue",
            highlight=True,
            style="blue",
        )
        if self._existing_topics:
            for topic in self._existing_topics:
                icon = "🔐 " if topic.startswith("__") else "📁 "
                tree.add(Text(icon) + f" {topic}")

        return Panel(
            Align.center(
                RenderGroup(tree),
                vertical="middle",
            ),
            title="[b blue] All available topics",
            border_style="bright_blue",
            box=box.ROUNDED,
            padding=(1, 2),
        )


class RichConsole:
    def __init__(
        self,
        topics: List[str],
        message_buffer: Dict[str, OrderedDict],
        existing_topics: List[str],
    ):
        self._layout = self.make_layout(topics)
        self._layout["header"].update(Header())
        self._layout["topics_tree"].update(TopicsTreePanel(existing_topics))
        for topic in topics:
            self._layout[topic].update(MessagePanel(topic, message_buffer[topic]))

    def update_console(self):
        with Live(self._layout, screen=True, refresh_per_second=1):
            while True:
                pass

    @staticmethod
    def make_layout(topics: List[str]) -> Layout:
        """Define the layout."""
        layout = Layout(name="root")

        layout.split(
            Layout(name="header", size=3),
            Layout(name="main", ratio=1),
        )
        layout["main"].split_row(
            Layout(name="topics_tree"),
            Layout(name="updates", ratio=2, minimum_size=60),
        )
        args = [Layout(name=topic) for topic in topics]
        layout["updates"].split(*args)
        return layout


class NormalConsole:
    def __init__(self, message_buffer: Dict[str, OrderedDict], logger: Logger):
        self._message_buffer = message_buffer
        self._logger = logger

    def update_console(self):
        while True:
            if self._message_buffer:
                for _, buffer in self._message_buffer.items():
                    if buffer:
                        items = buffer.popitem(last=False)
                        self._logger.error(items[1])



from getpass import getuser
from logging import Logger
from typing import List

from ess_message_consumer.console_output import NormalConsole, RichConsole
from ess_message_consumer.consumer import EssMessageConsumer
from ess_message_consumer.topic_watchdog import TopicWatchDog
from ess_message_consumer.utils import broker_cli, cli_parser, get_logger


class Application:
    def __init__(
        self, broker: str, topics: List[str], logger: Logger, rich_console: bool = False
    ):

        self._ess_message_consumer = EssMessageConsumer(broker, topics, logger)
        message_buffer = self._ess_message_consumer.message_buffer

        self._topic_watchdog = TopicWatchDog(broker, logger)
        existing_topics = self._topic_watchdog.existing_topics

        if rich_console:
            self._console = RichConsole(
                topics, message_buffer, existing_topics
            )  # type: ignore
        else:
            self._console = NormalConsole(message_buffer, logger)  # type: ignore

    def start(self):
        self._topic_watchdog.track_topics()
        self._ess_message_consumer.subscribe()
        self._console.update_console()

    def stop(self):
        for consumer in self._ess_message_consumer.consumers:
            consumer.close()


def start_application():
    args = cli_parser()

    topics = [x.strip() for x in args.topics.split(",") if x.strip()]
    broker = args.broker
    rich_console = args.rich_console

    logger = get_logger("ess-message-consumer", rich_console)

    app = Application(broker, topics, logger, rich_console=rich_console)

    try:
        app.start()
    except KeyboardInterrupt:
        logger.info(f"Interrupted by user: {getuser()}. Closing consumers ...")
    finally:
        app.stop()


def list_available_topics():
    args = broker_cli()
    broker = args.broker

    topic_watchdog = TopicWatchDog(broker)
    topics = topic_watchdog.get_available_topics()
    to_print = ""
    for index, value in enumerate(topics):
        to_print += f"  {index} : {value}\n"
    print(to_print)


if __name__ == "__main__":
    start_application()



import json

from streaming_data_types import (
    deserialise_6s4t,
    deserialise_ADAr,
    deserialise_answ,
    deserialise_ev42,
    deserialise_f142,
    deserialise_hs00,
    deserialise_pl72,
    deserialise_wrdn,
    deserialise_x5f2,
)


class StatusMessage:
    def deserialize(self, message):
        return deserialise_x5f2(message)


class FwResponseMessage:
    def deserialize(self, message):
        return deserialise_answ(message)


class FwFinishedMessage:
    def deserialize(self, message):
        return deserialise_wrdn(message)


class RunStopMessage:
    def deserialize(self, message):
        return deserialise_6s4t(message)


class RunStartMessage:
    def deserialize(self, message):
        return deserialise_pl72(message)


class LogMessage:
    def deserialize(self, message):
        return deserialise_f142(message)


class EventMessage:
    def deserialize(self, message):
        return deserialise_ev42(message)


class HistogramMessage:
    def deserialize(self, message):
        return deserialise_hs00(message)


class JsonMessage:
    def deserialize(self, message):
        return json.loads(message)


class AreaDetectorMessage:
    def deserialize(self, message):
        return deserialise_ADAr(message)


class DeserializerFactory:
    _message_handler = {
        b"x5f2": StatusMessage,
        b"answ": FwResponseMessage,
        b"wrdn": FwFinishedMessage,
        b"6s4t": RunStopMessage,
        b"pl72": RunStartMessage,
        b"f142": LogMessage,
        b"ev42": EventMessage,
        b"hs00": HistogramMessage,
        b"ADAr": AreaDetectorMessage,
        b"json": JsonMessage,
    }

    @classmethod
    def from_serialized_type(cls, type):
        if type in cls._message_handler:
            return cls._message_handler[type]()

        return cls._message_handler[b"json"]()



import time
import uuid
from logging import Logger
from typing import Dict, List

from confluent_kafka import Consumer  # type: ignore

from ess_message_consumer.deserializer import DeserializerFactory
from ess_message_consumer.utils import BoundOrderedDict, run_in_thread


class EssMessageConsumer:
    def __init__(self, broker: str, topics: List[str], logger: Logger):
        self._broker = broker
        self._topics = topics
        self._logger = logger

        self._message_buffer: Dict[str, BoundOrderedDict] = {
            topic: BoundOrderedDict(maxlen=50) for topic in self._topics
        }

        self._consumers = {}
        try:
            for topic in self._topics:
                conf = {
                    "bootstrap.servers": self._broker,
                    "auto.offset.reset": "latest",
                    "group.id": uuid.uuid4(),
                }
                self._consumers[topic] = Consumer(conf)
        except Exception as error:
            self._logger.error(f"Unable to create consumers: {error}")
            raise

    @property
    def consumers(self):
        return self._consumers.values()

    @property
    def message_buffer(self):
        return self._message_buffer

    def subscribe(self):
        if not self._topics:
            self._logger.error("Empty topic list")
            return

        for topic, consumer in self._consumers.items():
            # Remove all the subscribed topics
            consumer.unsubscribe()
            existing_topics = consumer.list_topics().topics

            if topic not in existing_topics:
                self._logger.error(
                    f"Provided topic {topic} does not exist. \n"
                    f"Available topics are {list(existing_topics.keys())}"
                )
                consumer.close()
                continue

            consumer.subscribe([topic])
            self._consume(topic)

    @run_in_thread
    def _consume(self, topic: str):
        while True:
            time.sleep(1)
            msg = self._consumers[topic].poll(1)
            if msg is None:
                continue
            if msg.error():
                self._logger.error(f"Error: {msg.error()}")
            else:
                value = msg.value()
                topic = msg.topic()
                type = value[4:8]
                try:
                    deserialized_message = DeserializerFactory.from_serialized_type(
                        type
                    ).deserialize(value)
                    self._update_message_buffer(topic, deserialized_message)
                except Exception as error:
                    self._logger.error(
                        f"Unrecognized serialized type {type}: message: {value}"
                        f"\nError:{error}"
                    )

    def _update_message_buffer(self, topic, value):
        self._message_buffer[topic][time.time()] = value



import argparse
import logging
import time
from collections import OrderedDict
from functools import wraps
from threading import Thread

from confluent_kafka import Producer  # type: ignore
from confluent_kafka.admin import AdminClient  # type: ignore
from rich.logging import RichHandler


class BoundOrderedDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        self.maxlen = kwargs.pop("maxlen", None)
        super().__init__(*args, **kwargs)
        self._maintain_length()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._maintain_length()

    def _maintain_length(self):
        if self.maxlen is not None:
            while len(self) > self.maxlen:
                self.popitem(last=False)


class Profiler:
    """Profiler class
    Usage:
        with Profiler(label="some name"):
            do_something()
    """

    def __init__(self, label="", enabled=True):
        self._label = label
        self._enabled = enabled
        self._start = None

    def __enter__(self):
        self._start = time.perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        if self._enabled:
            duration = time.perf_counter() - self._start
            print(f"Time for Execution of {self._label}: {duration * 1000:.4f} ms.")


def run_in_thread(original):
    @wraps(original)
    def wrapper(*args, **kwargs):
        t = Thread(target=original, args=args, kwargs=kwargs, daemon=True)
        t.start()
        return t

    return wrapper


def get_logger(
    name: str, level: int = logging.DEBUG, rich_console: bool = False
) -> logging.Logger:
    logger = logging.getLogger(name)
    if rich_console:
        console_handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(filename)s - %(levelname)s - %(message)s"
        )
        console_handler.setFormatter(formatter)
    else:
        console_handler = RichHandler(show_level=False, show_path=False)

    logger.addHandler(console_handler)
    logger.setLevel(level)

    return logger


def check_kafka_connection(broker_url: str):
    kafka_ready = False
    msg = None
    if ":" not in broker_url:
        msg = f"Unable to parse URL {broker_url}, should be of form localhost:9092"
        return kafka_ready, msg

    conf = {"bootstrap.servers": broker_url}
    try:
        producer = Producer(conf)
    except Exception as error:
        msg = error
        return kafka_ready, msg

    def delivery_callback(err, msg):
        nonlocal kafka_ready
        if not err:
            kafka_ready = True

    n_polls = 0
    while n_polls < 5 and not kafka_ready:
        producer.produce(
            "__waitUntilUp", value="Check if up", on_delivery=delivery_callback
        )
        producer.poll(6)
        n_polls += 1

    msg = (
        "Kafka up .."
        if kafka_ready
        else f"Cannot connect to broker {broker_url} in 30 secs."
    )

    if kafka_ready:
        admin = AdminClient(conf)
        futures = admin.delete_topics(["__waitUntilUp"])
        for topic, future in futures.items():
            try:
                future.result()
            except Exception as error:
                msg = f"failed to delete topic {topic}: {error}"

    return kafka_ready, msg


def cli_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="ESS Message consumer")
    parser.add_argument(
        "-t",
        "--topics",
        required=True,
        type=str,
        help="List of topics to consume messages from",
    )

    parser.add_argument(
        "-b",
        "--broker",
        type=str,
        default="localhost:9092",
        help="Kafka broker address",
    )
    parser.add_argument(
        "--rich_console", action="store_true", help="To get rich layout"
    )

    return parser.parse_args()


def broker_cli() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="ESS Message consumer")
    parser.add_argument(
        "-b",
        "--broker",
        type=str,
        default="localhost:9092",
        help="Kafka broker address",
    )
    return parser.parse_args()



import json
import time
from getpass import getuser
from logging import Logger

from confluent_kafka import Producer

from ess_message_consumer.utils import broker_cli, run_in_thread


class EssMessageProducer:
    def __init__(self, broker: str, logger: Logger = None):
        self._broker = broker
        self._producer = Producer({"bootstrap.servers": self._broker})

        self._logger = logger.info if logger else print
        self._stop = False
        self._poller_t = self._start_polling_producer()

    @run_in_thread
    def _start_polling_producer(self):
        while not self._stop:
            self._producer.poll(0.5)

    def close(self):
        self._stop = True
        self._poller_t.join()
        self._producer.flush(2)

    def produce(self, topic: str, payload: bytes, timestamp: int):
        def ack(err, _):
            if err:
                self._logger(f"Message failed delivery: {err} \n")

        try:
            self._producer.produce(topic, payload, on_delivery=ack, timestamp=timestamp)
        except Exception as error:
            self._logger(f"Message failed delivery: {error} \n")
        self._producer.poll(0)


def _get_topic_and_message():
    topic = input("Topic name: ")
    message = input("Message to send: ")
    return topic, message


def _to_exit():
    exit = False
    if input("Exit? (y/n): ").lower() in ["yes", "y"]:
        exit = True
    return exit


def start_producer():
    args = broker_cli()
    broker = args.broker

    ess_message_producer = EssMessageProducer(broker)
    try:
        while True:
            topic, message = _get_topic_and_message()
            ess_message_producer.produce(topic, json.dumps(message), int(time.time()))
            if _to_exit():
                ess_message_producer.close()
                break
    except KeyboardInterrupt:
        print(f"Interrupted by user: {getuser()}. Closing producer ...")
    finally:
        ess_message_producer.close()


if __name__ == "__main__":
    start_producer()



import os.path as osp
import re
import sys

from setuptools import find_packages, setup


def find_version():
    with open(osp.join("ess_message_consumer", "__init__.py"), "r") as f:
        match = re.search(r'^__version__ = "(\d+\.\d+\.\d+)"', f.read(), re.M)
        if match is not None:
            return match.group(1)
        raise RuntimeError("Unable to find version string.")


def get_readme_content():
    basedir = osp.dirname(osp.realpath(sys.argv[0]))
    with open(osp.join(basedir, "README.md"), "r") as f:
        content = f.read()
    return content


setup(
    name="ess-message-consumer",
    version=find_version(),
    author="Ebad Kamil",
    author_email="ebad.kamil@ess.eu",
    maintainer="Ebad Kamil",
    description="Kafka consumer to handle ESS flatbuffer messages and provide a rich display on console.",
    long_description=get_readme_content(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "start_consumer = ess_message_consumer.application:start_application",
            "start_producer = ess_message_consumer.producer:start_producer",
            "list_available_topics = ess_message_consumer.application:list_available_topics",
        ],
    },
    install_requires=[
        "confluent_kafka >= 1.7.0",
        "ess-streaming_data_types>=0.10.0",
        "rich>=10.6.0",
        "black==20.8b1",
        "flake8==3.8.4",
        "isort==5.7.0",
    ],
    extras_require={
        "test": [
            "pytest",
        ]
    },
    python_requires=">=3.6",
)



import time
import uuid
from logging import Logger

from confluent_kafka import Consumer

from ess_message_consumer.utils import run_in_thread


class TopicWatchDog:
    def __init__(self, broker: str, logger: Logger = None):
        self._broker = broker
        self._logger = logger.error if logger else print
        self._existing_topics = set()

        try:
            conf = {
                "bootstrap.servers": self._broker,
                "auto.offset.reset": "latest",
                "group.id": uuid.uuid4(),
            }
            self._consumer = Consumer(conf)
        except Exception as error:
            msg = f"Unable to create consumers: {error}"
            self._logger(msg)
            raise

    @run_in_thread
    def track_topics(self):
        while True:
            existing_topics = self.get_available_topics()
            self._existing_topics.update(existing_topics)
            time.sleep(5)

    def get_available_topics(self):
        return list(self._consumer.list_topics().topics.keys())

    @property
    def existing_topics(self):
        return self._existing_topics
